/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2012                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include <vcg/complex/complex.h>
#include <vcg/complex/append.h>
#include <wrap/io_trimesh/import_ply.h>
#include <wrap/io_trimesh/export_ply.h>
#include <vcg/complex/algorithms/update/color.h>
#include <vcg/complex/algorithms/clean.h>

using namespace vcg;
using namespace std;


class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public vcg::UsedTypes<	vcg::Use<MyVertex>   ::AsVertexType,
                                        vcg::Use<MyEdge>     ::AsEdgeType,
                                        vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex<MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::Color4b, vcg::vertex::BitFlags  >{};
class MyFace    : public vcg::Face< MyUsedTypes, vcg::face::FFAdj,  vcg::face::VertexRef, vcg::face::Color4b, vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge<MyUsedTypes>{};
class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> , std::vector<MyEdge>  > {};

/// \brief Simply offset the position of a vertex along the vertex normal.
/// It assumes Normalized Normals.
Point3f OffsetVertex(MyVertex &v, float amount)
{
  return v.P()+v.N()*amount;
}

/*! \brief Smooth the line of boundary between selected and unselected faces.

 Smoothing is performed by averaging the position of a vertex with the position of the adjacent ones on the boundary line.
 To simplify the managment of the line we topologically cut the mesh by detaching all the selected face adjacent to unselected faces.
 We use an attribute for storing the average position of the vertex.
 */
void SmoothSelectionBoundary(MyMesh &m)
{
  // Topological cut of the mesh alogn the selection boundary
  for(MyMesh::FaceIterator fi = m.face.begin();fi!= m.face.end();++fi)
  {
    for(int i=0;i<3;++i)
    {
      if(fi->IsS()!=fi->FFp(i)->IsS())
        face::FFDetachManifold(*fi,i);
    }
  }

  MyMesh::PerVertexAttributeHandle<Point3f> avgPosH =
      vcg::tri::Allocator<MyMesh>::AddPerVertexAttribute<Point3f> (m,std::string("avg"));
  vcg::tri::UpdateFlags<MyMesh>::VertexClearV(m);
  for(MyMesh::FaceIterator fi = m.face.begin();fi!= m.face.end();++fi)
  {
    for(int i=0;i<3;++i)
    {
      if(vcg::face::IsBorder(*fi,i) && !fi->V(i)->IsV())
      {
        fi->V(i)->SetV();
        vcg::face::Pos<MyFace> pos(&*fi,i,fi->V(i));
        Point3f sum(0,0,0);
        assert(pos.IsBorder());
        sum+=pos.VFlip()->P();
        sum+=pos.V()->P()*2.0f;  // Central vertex is weighted twice
        pos.NextB();
        sum+=pos.V()->P();
        assert(pos.IsBorder());
        avgPosH[fi->V(i)] = sum /4.0f;
      }
    }
  }

  for(MyMesh::VertexIterator vi = m.vert.begin();vi!= m.vert.end();++vi)
  {
    if(vi->IsV())
      vi->P()= avgPosH[vi];
  }

   vcg::tri::Allocator<MyMesh>::DeletePerVertexAttribute<Point3f> (m,avgPosH);
}


/*! \brief Construct a smooth labyrinth over the surface of a mesh
  Done in three steps:
  1) first select a subset of the faces of the mesh with the topology of a labyrint using a random topoogical visit
  2) smooth the boundary between selected and unselected faces
  3) Detach the selected faces, offset them along the surface normal and build the connecting walls
*/
int main(int argc, char *argv[])
{
  if(argc<3)
    {
    printf("\nUsage: FGT_20120126 <mesh.ply> <offsetPerc>\n"
           " <offsetPerc> is the fraction of the bbox diag used to offset (larger values smaller offset).\n Suggested value: 20..200\n\n");
      return -1;
    }

  MyMesh m;
  tri::io::ImporterPLY<MyMesh>::Open(m,argv[1]);
  printf("Read mesh %i %i\n",m.FN(),m.VN());

  float scaleFac=m.bbox.Diag()/atoi(argv[2]);

  tri::UpdateColor<MyMesh>::PerFaceConstant(m,Color4b::White);
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  tri::UpdateFlags<MyMesh>::FaceClearS(m);
  tri::UpdateFlags<MyMesh>::VertexClearV(m);
  int visitedCnt=0;
  std::vector<MyFace*> faceVec;
  faceVec.push_back(&(m.face[0]));
  faceVec.back()->SetS();

  // Random order triangle-by-triangle topological visit of a mesh.
  while(!faceVec.empty())
  {
    int randPos = rand()%faceVec.size();
    std::swap(faceVec[randPos],faceVec.back());
    MyFace *fp=faceVec.back();
    faceVec.pop_back();
    for(int i=0;i<3;i++)
    {
      if(!fp->FFp(i)->IsS())
      {
        MyFace *ffp = fp->FFp(i);
        bool badFace=false;
        int visitedVertexCnt=0;
        for(int j=0;j<3;j++)
        {
          if(ffp->FFp(j)->IsS())
            if(ffp->FFp(j) != fp) badFace =true;
          if(ffp->V(j)->IsV()) visitedVertexCnt++;
        }
        if(badFace==false && visitedVertexCnt<=2)
        {
          faceVec.push_back(ffp);
          ffp->SetS();
          for(int j=0;j<3;j++)
            ffp->V(j)->SetV();
          ffp->C()=Color4b::Red;
        }
      }
    }
    visitedCnt++;
  }

  MyMesh topMesh;
  tri::UpdateBounding<MyMesh>::Box(m);
  tri::UpdateNormal<MyMesh>::PerVertexAngleWeighted(m);
  tri::UpdateNormal<MyMesh>::NormalizePerVertex(m);

  // We have just completed the selection of the faces  of the 'labyrinth'
  // Now we smooth that boundaries before the deletion/extrusion
  SmoothSelectionBoundary(m);

  tri::Append<MyMesh,MyMesh>::Selected(topMesh,m);

  for(int i=0;i<topMesh.VN();++i)
    topMesh.vert[i].P() = OffsetVertex(topMesh.vert[i],scaleFac);

  MyMesh::FaceIterator fi;
  for(fi = m.face.begin();fi!= m.face.end();++fi)
  {
    if(fi->IsS())
      tri::Allocator<MyMesh>::DeleteFace(m,*fi);
  }
  vcg::tri::UpdateColor<MyMesh>::PerVertexConstant(m,Color4b::Gray);
  vcg::tri::UpdateColor<MyMesh>::PerVertexConstant(topMesh,Color4b::LightGray);

  MyMesh wallMesh;
  vcg::tri::Allocator<MyMesh>::CompactFaceVector(m);
  vcg::tri::UpdateTopology<MyMesh>::FaceFace(m);

  for(fi = m.face.begin();fi!= m.face.end();++fi)
  {
    for(int i=0;i<3;++i)
    {
      // for each edge on the border we will add 4 vertices and 2 triangles
      // to the wall mesh
      if(vcg::face::IsBorder(*fi,i))
      {
        MyMesh::VertexPointer newViVec[4];
        MyMesh::VertexIterator newVi=tri::Allocator<MyMesh>::AddVertices(wallMesh,4);
        MyMesh::FaceIterator newFi=tri::Allocator<MyMesh>::AddFaces(wallMesh,2);
        newViVec[0]=&*newVi;  newVi->P() = fi->V0(i)->P(); newVi++;
        newViVec[1]=&*newVi;  newVi->P() = fi->P1(i); newVi++;
        newViVec[2]=&*newVi;  newVi->P() = OffsetVertex(*(fi->V0(i)),scaleFac); newVi++;
        newViVec[3]=&*newVi;  newVi->P() = OffsetVertex(*(fi->V1(i)),scaleFac); newVi++;

        newViVec[0]->C()=Color4b::Gray;
        newViVec[1]->C()=Color4b::Gray;
        newViVec[2]->C()=Color4b::LightGray;
        newViVec[3]->C()=Color4b::LightGray;

        newFi->V(0) = newViVec[0];
        newFi->V(1) = newViVec[2];
        newFi->V(2) = newViVec[1];
        newFi++;
        newFi->V(0) = newViVec[1];
        newFi->V(1) = newViVec[2];
        newFi->V(2) = newViVec[3];
      }
    }
  }

  // Now Merge everything

  vcg::tri::Append<MyMesh,MyMesh>::Mesh(m,topMesh);
  vcg::tri::Append<MyMesh,MyMesh>::Mesh(m,wallMesh);

  vcg::tri::Clean<MyMesh>::RemoveDuplicateVertex(m);

  printf("Visited %i faces\n",visitedCnt);
  tri::io::ExporterPLY<MyMesh>::Save(m,"labyrinth.ply",tri::io::Mask::IOM_FACECOLOR|tri::io::Mask::IOM_VERTCOLOR);

  return 0;
}
